# Climb Timer version 1.0.0 by Steven Hartin (freddukes)
# ./addons/eventscripts/climbtimer/climbtimer.py

#################################
#### DO NOT EDIT THIS FILE, #####
####    IF YOU DO, THINGS   #####
####       COULD BREAK      #####
#################################

# Import Modules
import es
import popuplib
import playerlib
import langlib
import gamethread
import vecmath
import services
import effectlib
#import cmdlib

import cPickle
import os
import time

# psyco, more unstable but a lot faster
import psyco
psyco.full()

# Globals
finished   = []
started    = []
effects    = []
mapDicts   = {}
players    = {}
currentMap = es.ServerVar('eventscripts_currentmap')

info = es.AddonInfo()
info.author      = "Steven Hartin (freddukes)"
info.name        = "Climb Timer"
info.basename    = "climbtimer"
info.version     = "1.0.0a"
info.url         = "http://addons.eventscripts.com/addons/view/%s"%info.basename
info.description = "Time how long to climb a map" 

es.ServerVar(info.basename, info.version, info.description + " made by " + info.author).makepublic()

text = lambda identifier, options = {}, lang = "en" : "No strings.ini found in ../addons/eventscripts/%(basename)s/%(basename)s.py" % {"basename" : info.basename} 

""" Load the dictionary """
dictPath = os.path.join( es.getAddonPath( info.basename ), "data.db" )
if os.path.isfile(dictPath):
    fileStream = open(dictPath, 'r')
    mapDicts   = cPickle.load(fileStream)
    fileStream.close()
    
""" Load the languages """
langPath = os.path.join( es.getAddonPath( info.basename ), "strings.ini" )
if os.path.isfile(langPath):
    text = langlib.Strings(langPath)
    
""" Set up an auth check service, to return whether or not the user is authed. """
if services.isRegistered('auth'):
    auth_service = services.use('auth')
    auth_service.registerCapability('climbtimer', auth_service.ADMIN)
    isAuthed = lambda x: auth_service.isUseridAuthorized(x, 'climbtimer')
else:
    isAuthed = lambda x: False
    
def load():
    """ If the script is loaded halfway through a load, ensure the delays start """
    if str(currentMap):
        es_map_start({'mapname':str(currentMap)})
    
def es_map_start(event_var):
    """ Clear the lists """
    del started[:]
    del finished[:]
    players.clear()
    """ Save the database """
    fileStream = open(dictPath, 'w')
    cPickle.dump(mapDicts, fileStream)
    fileStream.close()
    """ Restart the timer if there isn't a timer already """
    gamethread.cancelDelayed('climbtimer_checkloop')
    gamethread.cancelDelayed('climbtimer_hudloop')
    if event_var['mapname'] in mapDicts:
        if "endpos" not in mapDicts[event_var['mapname']]:
            return
        if "startpos" not in mapDicts[event_var['mapname']]:
            return
        gamethread.delayedname(0.1, 'climbtime_checkloop', checkLoop)
        hudLoop()
    
def player_disconnect(event_var):
    """ If a player disconnects early, clear up the delay list. """
    userid = int(event_var['userid'])
    if userid in started:
        started.remove(userid)
    if userid in finished:
        finished.remove(userid)
    if userid in players:
        del players[userid]
    
def player_say(event_var):
    """ Executed when a player talks. Run the chat commands """
    message = event_var['text'].strip()
    userid  = int(event_var['userid'])
    if message == "!climbtimer":
        """ Admin menu """
        if isAuthed(userid):
            adminPopup.send(userid)
            
    elif message.startswith("!wr"):
        """ 
        Display a popup with the top 10 times for that map, or if a map
        was present as an argument, display for that map.
        
        !wr // display current map records
        !wr <map name> // displays <map name>'s current records 
        """
        if message.count(" "):
            mapName = " ".join(message.split()[1:])
        else:
            mapName = str(currentMap)
        if mapName not in mapDicts: return
        if "startpos" not in mapDicts[mapName]: return
        if "endpos" not in mapDicts[mapName]: return
        worldRecord = popuplib.create("climbtimer_worldrecords")
        worldRecord.addline("[World Records : %s]\n%s" % (mapName, "-" * 30) )
        sortedList  = sorted(mapDicts[mapName], sortDict)[:-2]
        if sortedList:
            for index in xrange( min(10, len(sortedList) ) ):
                worldRecord.addline("->%s. %s - %s" % (index + 1, mapDicts[mapName][sortedList[index]]["name"], formatTime(mapDicts[mapName][sortedList[index]]["time"]) ) )
        else:
            worldRecord.addline("[No places recorded]")
        worldRecord.addline("-" * 30 + "\n0.Cancel")
        worldRecord.send(userid)
        
    elif message.startswith("!time"):
        """ 
        Similar to a rank command. Tell the user their time or another person's
        time on that current map.
        
        !time // displays their time for that map
        !time <userid/name/steamid> // displays the user's time for that map
        
        TODO: Add in an additional [map] argument and test for the last argument
              and see if it's a dict item. If so, then they want information for
              that map.
        """
        mapName = str(currentMap)
        if mapName not in mapDicts: return
        if "startpos" not in mapDicts[mapName]: return
        if "endpos" not in mapDicts[mapName]: return
        if message.count(" "):
            message = " ".join(message.split()[1:])
            target  = es.getuserid(message)
        else:
            target  = userid
        if not target:
            tell(userid, 'player not found', {'name':message})
            return
        steamid    = es.getplayersteamid(target)
        sortedList = sorted(mapDicts[mapName], sortDict)[:-2]
        if steamid in sortedList:
            position = sortedList.index(steamid) + 1
        dictObject = mapDicts[mapName][steamid]
        tokens = {}
        tokens['name'] = dictObject['name']
        tokens['time'] = formatTime(dictObject['time'])
        tokens['rank'] = position
        tokens['totalrank'] = len(sortedList)
        tell(userid, 'rank', tokens)
            
ignore = ("startpos", "endpos")
def sortDict(keyOne, keyTwo):
    """ 
    Sort steamid's by time taken. Returns a list from shortest time to longest time
    ignores the startpos and endpos items
    """
    if keyOne in ignore or keyTwo in ignore: return 1
    mapName = str(currentMap)
    if mapDicts[mapName][keyOne]["time"] > mapDicts[mapName][keyTwo]["time"]: return 1
    if mapDicts[mapName][keyOne]["time"] < mapDicts[mapName][keyTwo]["time"]: return -1
    return 0
            
def hudLoop():
    """ Constantly loop the HUD for players telling them their current time spent climbing """
    for player in es.getUseridList():
        if player not in started: continue
        if player in finished: continue
        if player not in players: continue
        timeLeft = formatTime(time.time() - players[player])
        es.usermsg('create', 'climbtimer', 'HintText')
        es.usermsg('write',  'byte',       'climbtimer', -1)
        es.usermsg('write',  'string',     'climbtimer', tell(player, 'timer', {'time':timeLeft}, False) )
        es.usermsg('send',   'climbtimer', player)
        es.usermsg('delete', 'climbtimer')
    gamethread.delayedname(1, 'climbtimer_hudloop', hudLoop)
            
def checkLoop():
    """ Iterate through a list of all the players and check their location """
    try:
        for player in es.getUseridList():
            if player not in finished:
                if player in started:
                    lowerVertex, upperVertex = mapDicts[str(currentMap)]["endpos"]
                    timeTaken = time.time() - players[player]
                    if vecmath.isbetweenRect(es.getplayerlocation(player), lowerVertex, upperVertex):
                        """ Player has finished... """
                        """ Add their time to the dictionary which will save """
                        steamid   = es.getplayersteamid(player)
                        overwrite = True
                        if steamid not in mapDicts[str(currentMap)]:
                            mapDicts[str(currentMap)][steamid] = {'time':timeTaken, 'name':es.getplayername(player)}
                        else:
                            if mapDicts[str(currentMap)][steamid]['time'] > timeTaken:
                                mapDicts[str(currentMap)][steamid]['time'] = timeTaken
                            else:
                                overwrite = False
                        if overwrite:
                            tell(player, "new record")
                            
                        """ Tell all users that he finished and append his ID to the finished list """
                        tokens    = {}
                        tokens['name'] = es.getplayername(player)
                        tokens['time'] = formatTime(timeTaken)
                        for tellplayer in es.getUseridList():
                            tell(tellplayer, "finish", tokens)
                        finished.append(player)                    
                else:
                    lowerVertex, upperVertex = mapDicts[str(currentMap)]["startpos"]
                    if vecmath.isbetweenRect(es.getplayerlocation(player), lowerVertex, upperVertex):
                        """ Player has started """
                        tokens = {}
                        tokens['name'] = es.getplayername(player)
                        for tellplayer in es.getUseridList():
                            tell(tellplayer, "started", tokens)
                        started.append(player)
                        players[player] = time.time()
        gamethread.delayedname(0.1, 'climbtime_checkloop', checkLoop)
    except KeyError:
        pass
            
def tell(userid, message, options = {}, tellMessage = True):
    """ Just a simple wrapper to send a multi language text or return it if tellMessage is False """
    message = text(message, options, playerlib.getPlayer(userid).get("lang") )
    if tellMessage:
        es.tell(userid, '#multi', '#green[Climb Timer] #default- #lightgreen%s' % message)
    else:
        return message
        
def formatTime(seconds):
    """ Format the time in H:M:S when seconds is given. """
    hours, minutes   = divmod(seconds, 3600)
    minutes, seconds = divmod(minutes, 60)
    return "%s:%02d:%02d" % (int(hours), minutes, seconds)
    
def effectLoop(userid, start, red = 0, green = 255, blue = 0):
    """ A method to loop the effects """
    effectlib.drawBox(start, es.getplayerlocation(userid), red=red, green=green, blue=blue, seconds=0.1)
    gamethread.delayedname(0.1, 'climbtimer_effects', effectLoop, (userid, start, red, green, blue) )
    
def adminPopupMenuselect(userid, choice, popupid):
    """
    Handles the menu select from the admin menu.
    
    Note: You have to staticly cast the getplayerlocation to a list so we can
          edit the items in it as getplayerlocation returns a tuple which can't
          be edited. 
    """
    mapName = str(currentMap)
    if choice == 1:
        if mapName not in mapDicts:
            mapDicts[mapName] = {}
        if "startpos" not in mapDicts[mapName]:
            mapDicts[mapName]['startpos'] = [ (0,0,0), (0,0,0) ]
        if userid not in effects:
            effects.append(userid)
            tell(userid, 'added start one')
            start = list(es.getplayerlocation(userid))
            start[2] -= 5
            mapDicts[mapName]["startpos"][0] = start
            effectLoop(userid, start)
        else:
            effects.remove(userid)
            gamethread.cancelDelayed('climbtimer_effects')
            tell(userid, 'added start two')
            mapDicts[mapName]["startpos"][1] = list(es.getplayerlocation(userid))
            mapDicts[mapName]["startpos"][1][2] -= 5
            if "endpos" in mapDicts[mapName]:
                gamethread.delayedname(0.1, 'climbtime_checkloop', checkLoop)
                hudLoop()
        adminPopup.send(userid)
            
    elif choice == 2:
        if mapName not in mapDicts:
            mapDicts[mapName] = {}
        if "endpos" not in mapDicts[mapName]:
            mapDicts[mapName]['endpos'] = [ (0,0,0), (0,0,0) ]
        if userid not in effects:
            effects.append(userid)
            tell(userid, 'added end one')
            start = list(es.getplayerlocation(userid))
            start[2] -= 5
            mapDicts[mapName]["endpos"][0] = start
            effectLoop(userid, start, 255, 0)
        else:
            effects.remove(userid)
            gamethread.cancelDelayed('climbtimer_effects')
            tell(userid, 'added end two')
            mapDicts[mapName]["endpos"][1] = list(es.getplayerlocation(userid))
            mapDicts[mapName]["endpos"][1][2] -= 5
            if "startpos" in mapDicts[mapName]:
                gamethread.delayedname(0.1, 'climbtime_checkloop', checkLoop)
                hudLoop()
        adminPopup.send(userid)
            
    elif choice == 3:
        if mapName in mapDicts:
            del mapDicts[mapName]
            gamethread.cancelDelayed('climbtime_checkloop')
        tell(userid, 'delete')    
        
adminPopup = popuplib.create("climbtimer_admin")
adminPopup.addline("Climb Timer Admin")
adminPopup.addline("-" * 30)
adminPopup.addline("->1. Add/Move the start position")
adminPopup.addline("->2. Add/Move the finish position")
adminPopup.addline("->3. Delete map positions and times")
adminPopup.addline("-" * 30)
adminPopup.addline("0. Close")
adminPopup.menuselectfb = adminPopupMenuselect